/*
	Copyright 2007 Coraid, Inc.

	Author:
	Sam Hopkins
	sah@coraid.com

	We define _FILE_OFFSET_BITS to get 64 bit pread/pwrite support.
	We define _GNU_SOURCE to obtain O_LARGEFILE, O_NOATIME.
*/

#define _FILE_OFFSET_BITS 64
#define _GNU_SOURCE
#include <pthread.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>
#include <sys/sysinfo.h>
#include "p9p.h"

#define dprint(...) if (!dflag) ; else fprintf(stderr, __VA_ARGS__)
#define VER "ddt-6"

enum {
	Suser,
	Snice,
	Ssystem,
	Sidle,
	Siowait,
	Sirq,
	Ssoftirq,
	Nstats,
};

enum {
	/* 4GB io default */
	Dbs= 256*1024,
	Dcount= 16*1024,
	Dssecs= 10,

	USER_HZ= 100,
	MAXPT= 64
};

char *snames[] = {
[Suser] 	= "user",
[Snice] 	= "nice",
[Ssystem] 	= "system",
[Sidle] 	= "idle",
[Siowait]	= "iowait",
[Sirq]		= "irq",
[Ssoftirq]	= "softirq"
};

pthread_t threads[MAXPT];
pthread_mutex_t iomutex = PTHREAD_MUTEX_INITIALIZER;
int nthreads = 1;
unsigned long long nextio;
unsigned long rdstats[Nstats];
unsigned long wrstats[Nstats];
unsigned long long rticks;
unsigned long long wticks;
unsigned long long begin;
unsigned long long end;
unsigned long long tickspersec;
char file[4096];
char *block;
static int iofd;
unsigned long long bs = Dbs;
unsigned long long count = Dcount;
unsigned long long total;	// can't init here, "integer overflow expression" (*TI)
int fflag, ncpu;
int Uflag;
int rflag;
int wflag;
int cflag;
int dflag;
int ssecs = Dssecs;

#ifdef UNDEFINED

// unused, kept here for ease of future reference

// from include/asm-x86_64/msr.h and include/asm-i386/msr.h, respectively
#define rdtscll64(val) do { \
     unsigned int __a,__d; \
     asm volatile("rdtsc" : "=a" (__a), "=d" (__d)); \
     (val) = ((unsigned long)__a) | (((unsigned long)__d)<<32); \
} while(0)
#define rdtscll32(val) \
     __asm__ __volatile__("rdtsc" : "=A" (val))

unsigned long long
tsc(void)
{
	unsigned long long n = 0;

	if (sizeof(void *) == 4) {
		rdtscll32(n);
	} else {
		rdtscll64(n);
	}
	return n;
}

#endif // UNDEFINED

void
sysfatal(char *fmt, ...)
{
	va_list arg;

	va_start(arg, fmt);
	vfprintf(stderr, fmt, arg);
	va_end(arg);
	exit(1);
}

static char*
qtoken(char *s, char *sep)
{
	int quoting;
	char *t;

	quoting = 0;
	t = s;	/* s is output string, t is input string */
	while (*t != 0 && (quoting || strchr(sep, *t) == nil)) {
		if (*t != '\'') {
			*s++ = *t++;
			continue;
		}
		/* *t is a quote */
		if (!quoting) {
			quoting = 1;
			t++;
			continue;
		}
		/* quoting and we're on a quote */
		if (t[1] != '\'') {
			/* end of quoted section; absorb closing quote */
			t++;
			quoting = 0;
			continue;
		}
		/* doubled quote; fold one quote into two */
		t++;
		*s++ = *t++;
	}
	if (*s != '\0') {
		*s = '\0';
		if(t == s)
			t++;
	}
	return t;
}

int
tokenize(char *s, char **args, int maxargs)
{
	int nargs;
	static char qsep[] = " \t\r\n";

	for (nargs=0; nargs<maxargs; nargs++) {
		while (*s != '\0' && strchr(qsep, *s))
			s++;
		if (*s == '\0')
			break;
		args[nargs] = s;
		s = qtoken(s, qsep);
	}

	return nargs;
}

void
procstat(unsigned long stats[Nstats])
{
	int i, n, fd, argc;
	char buf[128], *p, *argv[16], **args;

	args = argv;
	fd = open("/proc/stat", O_RDONLY);
	if (fd < 0)
		sysfatal("procstat open failure: %s.\n", strerror(errno));
	n = read(fd, buf, sizeof buf - 1);
	if (n < 0)
		sysfatal("procstat read failure: %s.\n", strerror(errno));
	buf[n] = '\0';
	p = strchr(buf, '\n');
	if (p == nil)
		sysfatal("procstat can't find eol in '%s'.\n", buf);
	*p = '\0';
	argc = tokenize(buf, argv, nelem(argv));
	if (argc < Nstats + 1)
		sysfatal("procstat format assumption failure.\n");
	args++;
	for (i=0; i<Nstats; i++)
		stats[i] = atoi(args[i]) - stats[i];
	close(fd);
}

void *
iothread(void *vp)
{
	unsigned long long i;
	long long n;
	ssize_t (*rw)(int, void *, size_t, off_t) = vp;

	for (;;) {
		pthread_mutex_lock(&iomutex);
		i = nextio++;
		pthread_mutex_unlock(&iomutex);
		if (i >= count)
			break;
		n = rw(iofd, block, bs, bs*i);
		if (n != bs)
			sysfatal("io error %lld/%lld at offset %lld\n", n, bs, bs*i);
	}
	return 0;
}

void
threadio(ssize_t (*rw)(int, void *, size_t, off_t))
{
	int n, i;

	nextio = 0;
	if (nthreads <= 1) {
		iothread(rw);
		return;
	}
	for (i=0; i<nthreads; i++) {
		n = pthread_create(&threads[i], 0, iothread, (void *) rw);
		if (n)
			fprintf(stderr, "error creating thread %d\n", i);
		else
			dprint("created thread %d:%ld\n", i, threads[i]);
	}
	for (i=0; i<nthreads; i++) {
		dprint("joining thread %d:%ld\n", i, threads[i]);
		pthread_join(threads[i], 0);
	}
}

void
ddr(void)
{
	fprintf(stderr, "Reading from %s ... ", file);
	procstat(rdstats);
	threadio(&pread);
	procstat(rdstats);
	fprintf(stderr, "done.\n");
}

void
ddw(void)
{
	fprintf(stderr, "Writing to %s ... ", file);
	procstat(wrstats);
	threadio((ssize_t (*)(int, void *, size_t, off_t)) &pwrite);  // shut it!
	fprintf(stderr, "syncing ... ");
	sync();
	procstat(wrstats);
	fprintf(stderr, "done.\n");
}

void
raw(unsigned long stats[Nstats])
{
	int i;

	for (i=0; i<Nstats; i++)
		printf("%s=%ld ", snames[i], stats[i]);
	printf("\n");
}

void
cooked(unsigned long stats[Nstats])
{
	long long kbs, m, n;
	unsigned long cpuu;
	int i;

	for (i=0, n=0; i<Nstats; i++)
		n += stats[i];

	m = stats[Sidle] + stats[Siowait];
	kbs = (bs * count);
	kbs /= n;
	kbs *= USER_HZ * ncpu;
	kbs >>= 10;

	cpuu = (n - m) * 100 / n;

	printf("%6lld %6ld\n", kbs, cpuu);
}

void
ddt(void)
{
	printf("%-5lldMiB  KiB/s    CPU%% \n", (bs*count) >> 20LL);
	if (wflag) {
		printf("Write   ");
		cooked(wrstats);
	}
	if (rflag) {
		printf("Read    ");
		cooked(rdstats);
	}
	if (dflag) {
		printf("\n");
		printf("WRaw ");
		raw(wrstats);
		printf("RRaw ");
		raw(rdstats);
	}
}

int
countcpus(void)
{
	char buf[4096], *p, *r;
	int fd, n;

	fd = open("/proc/stat", O_RDONLY);
	if (fd < 0)
		sysfatal("cpus: can't open /proc/stat: %s\n", strerror(errno));
	n = read(fd, buf, sizeof buf - 1);
	if (n < 0)
		sysfatal("cpus: can't read /proc/stat: %s\n", strerror(errno));
	buf[n] = 0;

	n = -1;
	for (p=r=buf; (r=strchr(p, '\n')); p=r) {
		*r++ = '\0';
		if (strncmp(p, "cpu", 3) == 0)
			n++;
	}
	return n;
}

void
usage(void)
{
	fprintf(stderr, "usage: ddt [-?] [-c count] [-b bs] dir\n"); 
	exit(1);
}

void *
balloc(unsigned long bs)
{
	char *p;
	size_t ps;
	int i;

	ps = getpagesize();
	p = malloc(bs+ps);
	p += ps - 1;
	p -= (size_t) p % ps;

	for (i=0; i<bs; i++)
		p[i] = i;
	return p;
}

/* formatted str to ll */
unsigned long long
fstrtoll(char *str)
{
	unsigned long long n;
	char *p;

	n = strtoll(str, &p, 0);
	if (n < 0 || p == str)
		goto error;
	switch (*p) {
	case 'g':
	case 'G':
		n <<= 10;
	case 'm':
	case 'M':
		n <<= 10;
	case 'K':
	case 'k':
		n <<= 10;
		p++;
	}
	if (*p != '\0')
error:		sysfatal("invalid number format %s\n", str);
	return n;
}

/* check the total amount of I/O is greater than available memory */
void
memcheck(void)
{
	struct sysinfo s;
	unsigned long long m;

	if (sysinfo(&s) < 0) {
		perror("sysinfo failure");
		return;
	}
	m = s.totalram;
	m *= s.mem_unit;
	if (m > total)
		fprintf(stderr, "* warning: total I/O < total ram - throughput reported may not reflect I/O path.\n");
}

int
main(int argc, char *argv[])
{
	char *dir;
	int n;
	struct stat s;

	ARGBEGIN {
	case 'U':
		Uflag++;
		break;
	case 'b':
		bs = fstrtoll(EARGF(usage()));
		break;
	case 'c':
		cflag++;
		count = fstrtoll(EARGF(usage()));
		break;
	case 'd':
		dflag++;
		break;
	case 'f':
		strcpy(file, EARGF(usage()));
		fflag++;
		break;
	case 'n':
		nthreads = strtol(EARGF(usage()), 0, 0);
		if (nthreads < 0 || nthreads > MAXPT)
			sysfatal("invalid thread count value [0-%d].\n", MAXPT);
		break;
	case 'r':
		rflag++;
		break;
	case 's':
		ssecs = strtol(EARGF(usage()), 0, 0);
		break;
	case 't':
		total = fstrtoll(EARGF(usage()));
		break;
	case 'v':
		printf("%s\n", VER);
		return 0;
	case 'w':
		wflag++;
		break;
	case '?':
		usage();
	} ARGEND

	if (!wflag) {
		if (!rflag)
			wflag = rflag = 1;
		else if (!fflag)
			sysfatal("error: -r must be used with -f.\n");
	}

	if (fflag == 0) {
		if (argc != 1)
			usage();
		dir = *argv;
		n = strlen(dir);
		if (dir[--n] == '/')
			dir[n] = '\0';
		snprintf(file, sizeof file, "%s/ddt.%d", dir, getpid());
		iofd = open(file, O_CREAT|O_RDWR|O_LARGEFILE|O_NOATIME, 0666);
		if (!Uflag)
			unlink(file);
	} else if (!wflag) {
		iofd = open(file, O_RDWR|O_LARGEFILE|O_NOATIME);
		if (iofd >= 0)
		if (!cflag) {
			fstat(iofd, &s);
			count = s.st_size / bs;
		}
	} else
		iofd = open(file, O_CREAT|O_RDWR|O_LARGEFILE|O_NOATIME, 0666);

	if (total == 0)		// -t was not used
		total = bs*count;
	else
		count = total / bs;

	memcheck();

	if (iofd == -1)
		sysfatal("can't open %s: %s.\n", file, strerror(errno));

	ncpu = countcpus();
	block = balloc(bs);
	if (block == nil)
		sysfatal("can't malloc %d size block for I/O.\n", bs);
	if (wflag) {
		ddw();
		if (ssecs > 0) {
			fprintf(stderr, "sleeping %d seconds ... ", ssecs);
			sleep(ssecs);
			printf("done.\n");
		}
	}
	if (rflag)
		ddr();
	ddt();
	return 0;
}

